@@ -1,6 +1,8 @@
 from posixpath import dirname
 
+from django.conf import settings
 from django.contrib import admin
+from django.core.cache import caches
 from django.core.exceptions import PermissionDenied
 from django.http import (
     HttpRequest,
@@ -12,6 +14,7 @@ from django.template.response import TemplateResponse
 from django.urls import path
 from django.utils.html import format_html
 
+from .forms import LinkBackForm, TestRulesForm
 from .models import Rule, RuleNetwork, RulePath, RuleRatelimit, RuleSource
 from .utils import RULE_ACTION, LockoutException, get_ip, parse_ipaddress
 
@@ -24,7 +27,26 @@ except ImportError:
 # Register your models here.
 
 
-class RuleSubMixin:
+class TestRulesMixin:
+    def changelist_view(self, request, extra_context=None):
+        default = {
+            "test_rules_form": TestRulesForm(
+                data={
+                    **request.session.get("iprestrict_test_rules_data", {}),
+                    "test_rules_form-link_back": request.path,
+                },
+                prefix="test_rules_form",
+            )
+        }
+        if extra_context:
+            default.update(extra_context)
+        return super().changelist_view(
+            request,
+            extra_context=default,
+        )
+
+
+class RuleSubMixin(TestRulesMixin):
     def get_urls(self):
         return [
             path("<str:object_id>/change/", self.redirect_change),
@@ -107,12 +129,19 @@ _position_template = """
 
 
 @admin.register(Rule)
-class RuleAdmin(admin.ModelAdmin):
-    list_display = ("position_short", "position_buttons", "action", "name")
+class RuleAdmin(TestRulesMixin, admin.ModelAdmin):
+    list_display = (
+        "position_short",
+        "position_buttons",
+        "action",
+        "name",
+        "methods",
+        "invert_methods",
+    )
     list_display_links = ("position_short",)
-    list_editable = ("name", "action")
+    list_editable = ("name", "action", "methods", "invert_methods")
     ordering = ("position",)
-    fields = ["name", "action"]
+    fields = ["name", "action", "methods", "invert_methods"]
     inlines = [
         RuleNetworkInlineAdmin,
         RuleSourceInlineAdmin,
@@ -134,6 +163,7 @@ class RuleAdmin(admin.ModelAdmin):
     def get_urls(self):
         return [
             path("test_rules/", self.test_rules),
+            path("clear_caches/", self.clear_iprestrict_caches),
             path(
                 "simulate_rules<path:test_path>",
                 self.simulate_rules,
@@ -143,6 +173,9 @@ class RuleAdmin(admin.ModelAdmin):
         ]
 
     def rule_move(self, request: HttpRequest, object_id):
+        if request.method != "POST":
+            # fail soft for browsers visiting this url
+            return HttpResponseRedirect("../../")
         direction = request.POST["rule_move_direction"]
         parent_path = dirname(request.path.rstrip("/"))
         try:
@@ -183,6 +216,7 @@ class RuleAdmin(admin.ModelAdmin):
             for rdict in ratelimits:
                 r = ratelimit.get_ratelimit(
                     request=request,
+                    # don't use action
                     action=ratelimit.Action.PEEK,
                     group=rdict["group"],
                     key=rdict["key"],
@@ -199,36 +233,74 @@ class RuleAdmin(admin.ModelAdmin):
     def test_rules(self, request):
         from django.contrib.messages import ERROR, INFO, SUCCESS, WARNING
 
-        test_ip = request.POST.get("test_ip", None)
-        if not test_ip:
-            test_ip = get_ip(request)
+        if request.method != "POST":
+            # fail soft for browsers visiting this url
+            return HttpResponseRedirect("../")
+        form = TestRulesForm(
+            request.POST,
+            prefix="test_rules_form",
+        )
+        if form.is_valid():
+            test_method = request.POST.get("test_method", None) or None
 
-        test_path = request.POST.get("test_path", None) or ""
-        if test_path:
-            rule_id = RulePath.objects.match_ip_and_path(test_ip, test_path)[0]
-            self.message_user(
-                request,
-                f"Parameters: ip: {test_ip} (parsed: {parse_ipaddress(test_ip)}), path: {test_path}",
-                level=INFO,
-            )
-        else:
-            rule_id = Rule.objects.match_ip(test_ip)[0]
-            self.message_user(
-                request,
-                f"Parameters: ip: {test_ip} (parsed: {parse_ipaddress(test_ip)})",
-                level=INFO,
-            )
-        if rule_id:
-            rule = Rule.objects.get(id=rule_id)
-            self.message_user(
-                request,
-                f"Matched rule: {rule.name}, action: {RULE_ACTION(rule.action).label}",
-                level=SUCCESS if rule.action == RULE_ACTION.allow.value else WARNING,
-            )
-        else:
-            self.message_user(request, "No rule matched", level=ERROR)
+            test_ip = form.cleaned_data.get("test_ip", None)
+            if not test_ip:
+                test_ip = get_ip(request)
 
-        return HttpResponseRedirect("../")
+            test_path = form.cleaned_data.get("test_path", None) or ""
+            if test_path:
+                rule_id = RulePath.objects.match_ip_and_path(
+                    ip=test_ip, path=test_path, method=test_method
+                )[0]
+                self.message_user(
+                    request,
+                    f"Parameters: ip: {parse_ipaddress(test_ip)}, path: {test_path}, method: {test_method or '-'}",
+                    level=INFO,
+                )
+            else:
+                rule_id = Rule.objects.match_ip(ip=test_ip, method=test_method)[0]
+                self.message_user(
+                    request,
+                    f"Parameters: ip: {parse_ipaddress(test_ip)}, method: {test_method or '-'}",
+                    level=INFO,
+                )
+            if rule_id:
+                rule = Rule.objects.get(id=rule_id)
+                self.message_user(
+                    request,
+                    f"Matched rule: {rule.name}, action: {RULE_ACTION(rule.action).label}",
+                    level=SUCCESS
+                    if rule.action == RULE_ACTION.allow.value
+                    else WARNING,
+                )
+            else:
+                self.message_user(request, "No rule matched", level=ERROR)
+
+        session_data = {}
+
+        for key in form.fields:
+            prefixed_key = f"test_rules_form-{key}"
+            if key != "link_back" and prefixed_key in form.data:
+                session_data[prefixed_key] = form.data[prefixed_key]
+        request.session["iprestrict_test_rules_data"] = session_data
+        return HttpResponseRedirect(form.cleaned_data.get("link_back", "../"))
+
+    def clear_iprestrict_caches(self, request):
+        if request.method != "POST":
+            # fail soft for browsers visiting this url
+            raise HttpResponseRedirect("../")
+        form = LinkBackForm(
+            request.POST,
+            prefix="test_rules_form",
+        )
+        if form.is_valid():
+            # set => deduplicating
+            for cache_name in {
+                getattr(settings, "IPRESTRICT_CACHE", "default"),
+                getattr(settings, "RATELIMIT_CACHE", "default"),
+            }:
+                caches[cache_name].clear()
+        return HttpResponseRedirect(f"{form.cleaned_data.get('link_back', '../')}")
 
     def changeform_view(self, request, object_id=None, form_url="", extra_context=None):
         try:
