@@ -1,8 +1,11 @@
+import logging
 import re
 import time
 from contextlib import ExitStack
 from functools import lru_cache
+from typing import Optional
 
+from asgiref.sync import sync_to_async
 from django.conf import settings
 from django.core.cache import caches
 from django.core.exceptions import ValidationError
@@ -13,21 +16,26 @@ from django.db.transaction import atomic
 from django.utils.module_loading import import_string
 
 from .utils import (
+    RATELIMIT_ACTION,
     RULE_ACTION,
     LockoutException,
     get_default_action,
     get_default_interval,
+    invertedset,
     parse_ipaddress,
     parse_ipnetwork,
 )
 from .validators import (
     validate_generator_fn,
+    validate_methods,
     validate_network,
+    validate_path,
     validate_rate,
     validate_ratelimit_key,
     validate_regex,
 )
 
+logger = logging.getLogger(__name__)
 # Create your models here.
 
 _empty = ()
@@ -63,8 +71,8 @@ class RuleManager(models.Manager):
             return
         if path:
             if (
-                RulePath.objects.match_ip_and_path(ip, path, remote=remote)[1]
-                == RULE_ACTION.deny.value
+                RulePath.objects.match_ip_and_path(ip=ip, path=path, remote=remote)[1]
+                == RULE_ACTION.deny
             ):
                 # otherwise the caches could cause access errors until server restart
                 if clear_caches_on_error:
@@ -72,7 +80,7 @@ class RuleManager(models.Manager):
                 raise LockoutException("would lock current user out")
 
         else:
-            if self.match_ip(ip, remote=remote)[1] == RULE_ACTION.deny.value:
+            if self.match_ip(ip=ip, remote=remote)[1] == RULE_ACTION.deny:
                 # otherwise the caches could cause access errors until server restart
                 if clear_caches_on_error:
                     self.clear_local_caches()
@@ -118,13 +126,11 @@ class RuleManager(models.Manager):
             ).exclude(action=RULE_ACTION.disabled.value)
         for obj in queryset.distinct():
             patterns[obj.id] = (
-                ["*"]
-                if not obj.has_networks and not obj.has_sources
-                else obj.get_processed_networks(),
-                RULE_ACTION(obj.action),
-                # annotated
-                not obj.has_networks and not obj.has_sources,
-                obj.get_ratelimit_dicts(),
+                ["*"] if obj._is_catchall() else obj.get_processed_networks(),  # 0
+                obj.get_methods(),  # 1
+                RULE_ACTION(obj.action),  # 2
+                obj._is_catchall(),  # 3
+                obj.get_ratelimit_dicts(),  # 4
             )
         return patterns
 
@@ -133,16 +139,28 @@ class RuleManager(models.Manager):
             self.clear_local_caches()
         return self._ip_matchers_local(generic=generic)
 
-    def match_ip(self, ip: str, rule_id=None, remote=True):
+    aip_matchers_local = sync_to_async(ip_matchers_local)
+
+    def match_ip(
+        self,
+        ip: str,
+        method: Optional[str] = None,
+        rule_id=None,
+        generic=False,
+        remote=True,
+    ):
         ip_address_user = parse_ipaddress(ip)
         if rule_id:
-            item = self.ip_matchers_local().get(rule_id, None)
+            item = self.ip_matchers_local(generic=generic).get(rule_id, None)
             if not item:
                 return None, get_default_action(), False, _empty
+            if method and method not in item[1]:
+                return None, get_default_action(), False, _empty
+
             for network in item[0]:
                 try:
                     if network == "*" or ip_address_user in network:
-                        return rule_id, *item[1:]
+                        return rule_id, *item[2:]
                 except TypeError:
                     pass
             if remote:
@@ -153,63 +171,78 @@ class RuleManager(models.Manager):
                     for network in remote_networks:
                         try:
                             if ip_address_user in network:
-                                return rule_id, *item[1:]
+                                return rule_id, *item[2:]
                         except TypeError:
                             pass
 
         else:
-            ip_matchers = self.ip_matchers_local()
+            ip_matchers = self.ip_matchers_local(generic=generic)
             ip_matchers_remote = (
                 RuleSource.objects.ip_matchers_remote(ip_matchers.keys())
                 if remote
                 else {}
             )
             for rule_id, item in ip_matchers.items():
+                if method and method not in item[1]:
+                    continue
                 for network in item[0]:
                     try:
                         if network == "*" or ip_address_user in network:
-                            return rule_id, *item[1:]
+                            return rule_id, *item[2:]
                     except TypeError:
                         pass
 
                 for network in ip_matchers_remote.get(str(rule_id), _empty):
                     try:
                         if ip_address_user in network:
-                            return rule_id, *item[1:]
+                            return rule_id, *item[2:]
                     except TypeError:
                         pass
         return None, get_default_action(), False, _empty
 
-    def match_all_ip(self, ip: str, remote=True, generic=False):
+    aip_match_ip = sync_to_async(match_ip)
+
+    def match_all_ip(
+        self, ip: str, method: Optional[str] = None, remote=True, generic=False
+    ):
         # note: cannot apply ratelimit here
         ip_address_user = parse_ipaddress(ip)
         result = []
         ip_matchers = self.ip_matchers_local(generic=generic)
-        ip_matchers_remote = (
-            RuleSource.objects.ip_matchers_remote(ip_matchers.keys()) if remote else {}
-        )
+        ip_matchers_remote = None
         for rule_id, item in ip_matchers.items():
+            if method and method not in item[1]:
+                continue
             found_local_network = False
             for network in item[0]:
                 try:
                     if network == "*" or ip_address_user in network:
-                        result.append((rule_id, *item[1:]))
+                        result.append((rule_id, *item[2:]))
                         found_local_network = True
                         break
                 except TypeError:
                     pass
             if found_local_network:
                 continue
+            # lazy fetch
+            if ip_matchers_remote is None:
+                ip_matchers_remote = (
+                    RuleSource.objects.ip_matchers_remote(ip_matchers.keys())
+                    if remote
+                    else {}
+                )
             for network in ip_matchers_remote.get(str(rule_id), _empty):
                 try:
                     if ip_address_user in network:
-                        result.append((rule_id, *item[1:]))
+                        result.append((rule_id, *item[2:]))
                         break
                 except TypeError:
                     pass
 
         return result
 
+    amatch_all_ip = sync_to_async(match_all_ip)
+
     def position_start(self, rule_id, ip=None, path=None):
         with self._atomic_update(ip=ip, path=path) as stack:
             stack.queryset.filter(id=rule_id).update(position=0)
@@ -258,6 +291,18 @@ class Rule(models.Model):
     _trigger_cleanup = True
     position = models.PositiveIntegerField(blank=True, default=get_default_position)
     name = models.CharField(max_length=50, unique=True)
+    methods = models.CharField(
+        max_length=100,
+        validators=[validate_methods],
+        default="",
+        blank=True,
+        help_text="comma seperated http methods",
+    )
+    invert_methods = models.BooleanField(
+        default=True,
+        blank=True,
+        help_text="exclude instead of include http methods. Set and leave methods empty to match all",
+    )
     action = models.CharField(
         max_length=1, choices=RULE_ACTION.choices, default=RULE_ACTION.disabled
     )
@@ -274,8 +319,12 @@ class Rule(models.Model):
         return self.name
 
     def is_catch_all(self, also_disabled=False):
-        rule = type(self).objects.ip_matchers_local(also_disabled).get(self.id)
-        return bool(rule and rule[2])
+        rule = type(self).objects.ip_matchers_local(generic=also_disabled).get(self.id)
+        return bool(rule and rule[3])
+
+    def _is_catchall(self):
+        # use annotations
+        return not self.has_networks and not self.has_sources
 
     def get_processed_networks(self):
         return [
@@ -288,12 +337,31 @@ class Rule(models.Model):
     def get_ratelimit_dicts(self):
         return list(
             self.ratelimits.filter(is_active=True).values(
-                "key", "group", "rate", "decorate_name", "block", "wait"
+                "key", "group", "rate", "decorate_name", "block", "wait", "action"
             )
         )
 
-    def match_ip(self, ip, remote=True):
-        return type(self).objects.match_ip(ip, rule_id=self.id, remote=remote)
+    def get_methods(self):
+        if self.invert_methods:
+            if not self.methods:
+                return invertedset()
+            return invertedset(self.methods.split(","))
+        if not self.methods:
+            return set()
+        return set(self.methods.split(","))
+
+    def match_ip(self, ip, method=None, remote=True, generic=False):
+        return type(self).objects.match_ip(
+            ip, method=method, rule_id=self.id, remote=remote, generic=generic
+        )
+
+    amatch_ip = sync_to_async(match_ip)
+
+    def clean(self):
+        super().clean()
+        self.methods = ",".join(
+            sorted(map(lambda x: x.upper().strip(), self.methods.split(",")))
+        )
 
 
 class RuleRatelimitManager(models.Manager):
@@ -310,6 +378,10 @@ class RuleRatelimit(models.Model):
     )
     group = models.CharField(max_length=50)
     decorate_name = models.CharField(max_length=50, default="ratelimit", blank=True)
+    action = models.SmallIntegerField(
+        choices=RATELIMIT_ACTION.choices,
+        default=RATELIMIT_ACTION.INCREASE,
+    )
     rate = models.CharField(max_length=10, validators=[validate_rate])
     block = models.BooleanField(blank=True, default=False)
     wait = models.BooleanField(blank=True, default=False)
@@ -328,10 +400,12 @@ class RuleNetwork(models.Model):
 
 
 class RuleSourceManager(models.Manager):
-    def clear_remote_caches(self):
+    def clear_remote_caches(
+        self,
+    ):
         cache = caches[getattr(settings, "IPRESTRICT_CACHE", "default")]
         cache.delete_many(
-            self.annotate_with_cache_key().values_list("cache_key", flat=True)
+            self.annotate_with_cache_key().values_list("cache_key", flat=True),
         )
 
     def annotate_with_cache_key(self, queryset=None):
@@ -347,9 +421,34 @@ class RuleSourceManager(models.Manager):
                 models.Value(":"),
                 models.F("rule_id"),
                 output_field=models.TextField(),
-            )
+            ),
+            force_expire_cache_key=Concat(
+                models.Value(
+                    f'{getattr(settings, "IPRESTRICT_KEY_PREFIX", "fip:")}source_force_expire:'
+                ),
+                models.F("interval"),
+                models.Value(":"),
+                models.F("rule_id"),
+                output_field=models.TextField(),
+            ),
         )
 
+    def _is_force_expired(self, query, cache, set_all):
+        if set_all:
+            return set(query.values_list("cache_key", flat=True))
+        expire_dates = cache.get_many(
+            set(query.values_list("force_expire_cache_key", flat=True))
+        )
+        cur_time = int(time.time())
+        expired = set()
+        for source in query:
+            if (
+                source.force_expire_cache_key not in expire_dates
+                or expire_dates[source.force_expire_cache_key] < cur_time
+            ):
+                expired.add(source.cache_key)
+        return expired
+
     def ip_matchers_remote(self, rules):
         cache = caches[getattr(settings, "IPRESTRICT_CACHE", "default")]
         q = models.Q(is_active=True)
@@ -359,26 +458,50 @@ class RuleSourceManager(models.Manager):
             q &= models.Q(rule_id__in=rules)
         keys_query = self.annotate_with_cache_key(self.filter(q))
         keys = keys_query.values_list("cache_key", flat=True)
-
-        # unordered dict
-        cache_result = cache.get_many(keys)
+        max_interval = keys_query.aggregate(Max("interval", default=0))["interval__max"]
+        force_expire = bool(getattr(settings, "IPRESTRICT_SOURCE_FORCE_EXPIRE", True))
+        # <= force_expire_multiplier <= 0 disables force expire
+        skip_cache = (
+            set()
+            if not force_expire
+            else self._is_force_expired(keys_query, cache, max_interval == 0)
+        )
+        cache_result = (
+            cache.get_many(set(keys).difference(skip_cache))
+            if max_interval > 0 and not skip_cache
+            else {}
+        )
         last_interval = None
         to_set = {}
         result = {}
-        for source in keys_query.exclude(cache_key__in=cache_result.keys()).order_by(
-            "interval"
+
+        for source in (
+            keys_query.exclude(cache_key__in=cache_result.keys())
+            .distinct()
+            .order_by("interval")
         ):
             # FIXME: double calling, some kind of locking would be good
             networks = source.get_processed_networks_uncached()
-            cache_key = source.get_cache_key()
+            # annotated
+            cache_key = source.cache_key
+            force_expire_cache_key = source.force_expire_cache_key
             result.setdefault(cache_key.rsplit(":", 1)[-1], []).extend(networks)
             if last_interval is not None and last_interval != source.interval:
-                cache.set_many(to_set, last_interval)
+                # not empty
+                if to_set:
+                    cache.set_many(to_set, timeout=last_interval)
                 to_set = {}
-            to_set[cache_key] = ",".join(map(lambda x: x.compressed, networks))
+            # interval 0 disables caching
+            if source.interval > 0:
+                to_set[cache_key] = ",".join(map(lambda x: x.compressed, networks))
+                if force_expire:
+                    to_set[force_expire_cache_key] = int(time.time()) + source.interval
+
             last_interval = source.interval
         if last_interval is not None:
-            cache.set_many(to_set, last_interval)
+            # not empty
+            if to_set:
+                cache.set_many(to_set, timeout=last_interval)
 
         for cache_key, value in cache_result.items():
             if isinstance(value, str):
@@ -412,6 +535,9 @@ class RuleSource(models.Model):
         # last is rule id for easy extraction
         return f'{getattr(settings, "IPRESTRICT_KEY_PREFIX", "fip:")}source_data:{self.id}:{self.rule_id}'
 
+    def get_force_expire_cache_key(self):
+        return f'{getattr(settings, "IPRESTRICT_KEY_PREFIX", "fip:")}source_force_expire:{self.interval}:{self.rule_id}'
+
     def get_processed_networks_uncached(self):
         ret = []
         success = False
@@ -423,12 +549,24 @@ class RuleSource(models.Model):
         if success:
             try:
                 ip_nets = import_string(self.generator_fn)()
-            except ImportError:
-                # FIXME: proper reporting
-                return
-            except Exception:
-                # FIXME: proper reporting
-                return
+            except ImportError as exc:
+                logger.warning(
+                    'could not import source generator_fn: "%s", rule name %s, source id %s',
+                    self.generator_fn,
+                    self.rule.name,
+                    self.id,
+                    exc_info=exc,
+                )
+                return []
+            except Exception as exc:
+                logger.error(
+                    'source generator_fn "%s" failed, rule name %s, source id %s',
+                    self.generator_fn,
+                    self.rule.name,
+                    self.id,
+                    exc_info=exc,
+                )
+                return []
             for network in ip_nets:
                 try:
                     ret.append(parse_ipnetwork(network))
@@ -450,24 +588,44 @@ class RulePathManager(models.Manager):
             patterns[key] = re.compile("|".join(patterns[key]))
         return patterns
 
-    def match_ip_and_path(self, ip: str, path: str, remote=True):
+    apath_matchers = sync_to_async(path_matchers)
+
+    def match_ip_and_path(
+        self,
+        ip: str,
+        path: str,
+        method: Optional[str] = None,
+        rule_id=None,
+        remote=True,
+    ):
         # ordered
         # with generic = pathless, non-catchall and disabled candidates are included
-        candidates = Rule.objects.match_all_ip(ip, remote=remote, generic=True)
+        if rule_id:
+            candidate = Rule.objects.match_ip(
+                ip, rule_id=rule_id, method=method, remote=remote, generic=True
+            )
+            if candidate[0] is not None:
+                candidates = [candidate]
+        else:
+            candidates = Rule.objects.match_all_ip(
+                ip, method=method, remote=remote, generic=True
+            )
         _path_matchers = self.path_matchers()
         ratelimits = []
-        for rule_id, action, is_catch_all, _ratelimits in candidates:
+        for _rule_id, action, is_catch_all, _ratelimits in candidates:
             if action == RULE_ACTION.disabled:
                 continue
-            matcher = _path_matchers.get(rule_id, None)
+            matcher = _path_matchers.get(_rule_id, None)
             # matcher is None, catchall for pathes (can be also a real catch all)
             if matcher and matcher.match(path):
                 ratelimits.extend(_ratelimits)
                 if action == RULE_ACTION.only_ratelimit:
                     continue
-                return rule_id, action, is_catch_all, ratelimits
+                return _rule_id, action, is_catch_all, ratelimits
         return None, get_default_action(), False, ratelimits
 
+    amatch_ip_and_path = sync_to_async(match_ip_and_path)
+
 
 class RulePath(models.Model):
     rule = models.ForeignKey(Rule, related_name="pathes", on_delete=models.CASCADE)
@@ -488,16 +646,10 @@ class RulePath(models.Model):
 
     def clean_fields(self, exclude=None):
         super().clean_fields(exclude=exclude)
-        if self.is_regex:
-            try:
-                validate_regex(self.path)
-            except ValidationError as exc:
-                if exclude and "path" in exclude:
-                    raise exc
-                else:
-                    raise ValidationError({"path": exc})
-                if exclude and "path" in exclude:
-                    raise exc
-                else:
-                    raise ValidationError({"path": exc})
-                    raise ValidationError({"path": exc})
+        try:
+            validate_regex(self.path) if self.is_regex else validate_path(self.path)
+        except ValidationError as exc:
+            if exclude and "path" in exclude:
+                raise exc
+            else:
+                raise ValidationError({"path": exc})
